<?php

namespace drp\DynamicAttributes;

use Illuminate\Database\Eloquent\Model;
use Schema;

/**
 * @property \Illuminate\Support\Collection $dynamicAttributes Dynamic
 * Attributes relationship
 *
 * @method \Illuminate\Database\Eloquent\Builder whereDynamicAttribute()
 * whereDynamicAttribute(string $field, string $operator, mixed $value)
 * Perform where query on one of the dynamic attributes
 */
class DynamicModel extends Model
{
    /**
     * All the fields that are currently stored in the key store database
     * tables.
     *
     * @var array
     */
    protected $dynamicFields = [];

    /**
     * All the dirty dynamic fields that will are to be updated.
     *
     * @var array
     */
    protected $dirtyDynamicFields = [];

    /**
     * New dynamic fields that will be stored.
     *
     * @var array
     */
    protected $newDynamicFields = [];

    /**
     * Schema of the dynamic attributes
     *
     * @var \Illuminate\Support\Collection
     */
    protected $schema;

    /**
     * The relations to eager load on every query.
     *
     * @var array
     */
    protected $with = [
        'dynamicAttributes',
    ];

    /**
     * Dynamic attributes relationship.
     *
     * @return \Illuminate\Database\Eloquent\Relations\HasMany
     */
    public function dynamicAttributes()
    {
        $relationship = $this->hasMany(Attribute::class, 'model_id', 'id');

        return $relationship;
    }

    /**
     * Fill the model with an array of attributes.
     *
     * @param  array $attributes Attributes to add to the model, if they are
     *                           fillable
     * @return $this
     *
     * @throws \Illuminate\Database\Eloquent\MassAssignmentException
     */
    public function fill(array $attributes)
    {
        parent::fill($attributes);
        $this->schema = AttributeSchema::where(
            'model',
            get_class($this)
        )->get();

        return $this;
    }

    /**
     * Set the array of model attributes. No checking is done.
     *
     * @param  array $attributes The attributes to set
     * @param  bool  $sync       Whether to sync attributes with original
     * @return $this
     */
    public function setRawAttributes(array $attributes, $sync = false)
    {
        parent::setRawAttributes($attributes, $sync);

        $dynamicAttributes = $this->dynamicAttributes;
        foreach ($dynamicAttributes as $attribute) {
            $this->setAttribute($attribute->field, $attribute->value);
            $this->dynamicFields[$attribute->field] = $attribute->value;
        }

        return $this;
    }

    /**
     * Save the model to the database.
     *
     * @param  array $options Options when saving
     * @return bool
     */
    public function save(array $options = [])
    {
        $this->validateAnyExistingValuesAgainstSchema();
        $this->cleanModelAttributesOfDynamicAttributes();

        $result = parent::save($options);
        $this->processDynamicFields();

        return $result;
    }

    /**
     * Validates any dynamic values that currently exist within attributes,
     * using the schema.
     *
     * @return null
     */
    protected function validateAnyExistingValuesAgainstSchema()
    {
        $modelSchema    = $this->schema->keyBy('field');
        $existingValues = array_intersect_key(
            $this->attributes,
            array_flip($modelSchema->keys()->toArray())
        );

        foreach ($existingValues as $field => $value) {
            $fieldSchema = $modelSchema->get($field);
            if (!$fieldSchema->allow_null && $value == null) {
                throw new SchemaValidationException(
                    'Column ' . $field . ' can not be null'
                );
            }

            if (gettype($value) != $fieldSchema->type) {
                throw new SchemaValidationException(
                    'Column ' . $field . ' must of type ' . $fieldSchema->type
                );
            }
        }
    }

    /**
     * Creates new attributes or updates old depending on what the
     * dirtyDynamicFields and newDynamicFields array have.
     *
     * @return null
     */
    protected function processDynamicFields()
    {
        $this->processExistingDynamicFields();
        $this->processNewDynamicFields();

        $this->load('dynamicAttributes');
        $this->setRawAttributes($this->attributes);
    }

     /**
     * Loops through any existing dynamic fields and updates the correct
     * attribute model. Then clears the dirtyDynamicFields array
     *
     * @return null
     */
    protected function processExistingDynamicFields()
    {
        $keyedSchema    = $this->schema->keyBy('field');
        $fieldsToUpdate = array_diff(
            $this->dirtyDynamicFields,
            $keyedSchema->keys()->toArray()
        );
        foreach ($fieldsToUpdate as $fieldName => $value) {
            $keyedAttributes = $this->dynamicAttributes->keyBy('field');
            $attribute       = $keyedAttributes->get($fieldName);
            if ($attribute === null) {
                $attribute            = new Attribute();
                $attribute->model_id  = $this->getKey();
                $attribute->schema_id = $keyedSchema->get($fieldName)->id;
            }

            $attribute->value = $value;
            $attribute->save();
        }

        $this->dirtyDynamicFields = [];
    }

    /**
     * Loops through any new dynamic fields and creates a new schema and
     * attribute models. Then clears the newDynamicFields array
     *
     * @return null
     */
    protected function processNewDynamicFields()
    {
        foreach ($this->newDynamicFields as $field => $value) {
            AttributeSchema::createNewSchemaAndValue($this, $field, $value);
        }

        $this->newDynamicFields = [];
    }

    /**
     * Moves any of the non model attributes from the attributes array.
     *
     * @return null
     */
    protected function cleanModelAttributesOfDynamicAttributes()
    {
        foreach ($this->schema->pluck('field') as $dynamicField) {
            if (array_key_exists($dynamicField, $this->attributes)) {
                $this->moveDynamicFieldFromAttributes($dynamicField);
            }
        }

        $modelColumns     = Schema::getColumnListing($this->getTable());
        $newDynamicFields = array_diff(
            array_keys($this->attributes),
            $modelColumns
        );

        foreach ($newDynamicFields as $dynamicField) {
            $this->moveDynamicFieldFromAttributes(
                $dynamicField,
                'newDynamicFields'
            );
        }
    }

    /**
     * Moves a field from the attributes array to the dirtyDynamicFields array.
     *
     * @param string $dynamicField The field to move
     * @param string $array        The array to move to
     *
     * @return null
     */
    protected function moveDynamicFieldFromAttributes(
        $dynamicField,
        $array = 'dirtyDynamicFields'
    ) {
        $this->{$array}[$dynamicField] = $this->attributes[$dynamicField];
        unset($this->attributes[$dynamicField]);
    }

    /**
     * Adds scope so you can do wheres on the dynamic attributes
     *
     * @param Builder $query    Query builder to manipulate
     * @param string  $field    The field to compare against
     * @param string  $operator The operator to use when comparing against value
     * @param mixed   $value    The value to compare
     *
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public function scopeWhereDynamicAttribute(
        $query,
        $field,
        $operator,
        $value
    ) {
        if ($operator === 'LIKE') {
            $value = '%' . $value . '%';
        }

        $query->whereHas(
            'dynamicAttributes',
            function ($query) use ($field, $operator, $value) {
                $query->whereHas(
                    'schema',
                    function ($query) use ($field) {
                        $query->where('field', '=', $field);
                    }
                );
                $query->where('value', $operator, $value);
            }
        );

        return $query;
    }
}
